// Copyright (c) 2014 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifndef CSSPropertyNames_h
#define CSSPropertyNames_h

#include "core/CoreExport.h"
#include "platform/wtf/Assertions.h"
#include <stddef.h>

namespace WTF {
class AtomicString;
class String;
}

namespace blink {

enum CSSPropertyID {
    CSSPropertyInvalid = 0,
    // This isn't a property, but we need to know the position of @apply rules in style rules
    CSSPropertyApplyAtRule = 1,
    CSSPropertyVariable = 2,
    CSSPropertyAnimationDelay = 3,
    CSSPropertyAnimationDirection = 4,
    CSSPropertyAnimationDuration = 5,
    CSSPropertyAnimationFillMode = 6,
    CSSPropertyAnimationIterationCount = 7,
    CSSPropertyAnimationName = 8,
    CSSPropertyAnimationPlayState = 9,
    CSSPropertyAnimationTimingFunction = 10,
    CSSPropertyTransitionDelay = 11,
    CSSPropertyTransitionDuration = 12,
    CSSPropertyTransitionProperty = 13,
    CSSPropertyTransitionTimingFunction = 14,
    CSSPropertyColor = 15,
    CSSPropertyDirection = 16,
    CSSPropertyFontFamily = 17,
    CSSPropertyFontFeatureSettings = 18,
    CSSPropertyFontKerning = 19,
    CSSPropertyFontSize = 20,
    CSSPropertyFontSizeAdjust = 21,
    CSSPropertyFontStretch = 22,
    CSSPropertyFontStyle = 23,
    CSSPropertyFontVariantCaps = 24,
    CSSPropertyFontVariantLigatures = 25,
    CSSPropertyFontVariantNumeric = 26,
    CSSPropertyFontVariationSettings = 27,
    CSSPropertyFontWeight = 28,
    CSSPropertyTextOrientation = 29,
    CSSPropertyTextRendering = 30,
    CSSPropertyWebkitFontSmoothing = 31,
    CSSPropertyWebkitLocale = 32,
    CSSPropertyWebkitTextOrientation = 33,
    CSSPropertyWebkitWritingMode = 34,
    CSSPropertyWritingMode = 35,
    CSSPropertyZoom = 36,
    CSSPropertyAlignContent = 37,
    CSSPropertyAlignItems = 38,
    CSSPropertyAlignSelf = 39,
    CSSPropertyAlignmentBaseline = 40,
    CSSPropertyAll = 41,
    CSSPropertyAnimation = 42,
    CSSPropertyBackdropFilter = 43,
    CSSPropertyBackfaceVisibility = 44,
    CSSPropertyBackground = 45,
    CSSPropertyBackgroundAttachment = 46,
    CSSPropertyBackgroundBlendMode = 47,
    CSSPropertyBackgroundClip = 48,
    CSSPropertyBackgroundColor = 49,
    CSSPropertyBackgroundImage = 50,
    CSSPropertyBackgroundOrigin = 51,
    CSSPropertyBackgroundPosition = 52,
    CSSPropertyBackgroundPositionX = 53,
    CSSPropertyBackgroundPositionY = 54,
    CSSPropertyBackgroundRepeat = 55,
    CSSPropertyBackgroundRepeatX = 56,
    CSSPropertyBackgroundRepeatY = 57,
    CSSPropertyBackgroundSize = 58,
    CSSPropertyBaselineShift = 59,
    CSSPropertyBlockSize = 60,
    CSSPropertyBorder = 61,
    CSSPropertyBorderBottom = 62,
    CSSPropertyBorderBottomColor = 63,
    CSSPropertyBorderBottomLeftRadius = 64,
    CSSPropertyBorderBottomRightRadius = 65,
    CSSPropertyBorderBottomStyle = 66,
    CSSPropertyBorderBottomWidth = 67,
    CSSPropertyBorderCollapse = 68,
    CSSPropertyBorderColor = 69,
    CSSPropertyBorderImage = 70,
    CSSPropertyBorderImageOutset = 71,
    CSSPropertyBorderImageRepeat = 72,
    CSSPropertyBorderImageSlice = 73,
    CSSPropertyBorderImageSource = 74,
    CSSPropertyBorderImageWidth = 75,
    CSSPropertyBorderLeft = 76,
    CSSPropertyBorderLeftColor = 77,
    CSSPropertyBorderLeftStyle = 78,
    CSSPropertyBorderLeftWidth = 79,
    CSSPropertyBorderRadius = 80,
    CSSPropertyBorderRight = 81,
    CSSPropertyBorderRightColor = 82,
    CSSPropertyBorderRightStyle = 83,
    CSSPropertyBorderRightWidth = 84,
    CSSPropertyBorderSpacing = 85,
    CSSPropertyBorderStyle = 86,
    CSSPropertyBorderTop = 87,
    CSSPropertyBorderTopColor = 88,
    CSSPropertyBorderTopLeftRadius = 89,
    CSSPropertyBorderTopRightRadius = 90,
    CSSPropertyBorderTopStyle = 91,
    CSSPropertyBorderTopWidth = 92,
    CSSPropertyBorderWidth = 93,
    CSSPropertyBottom = 94,
    CSSPropertyBoxShadow = 95,
    CSSPropertyBoxSizing = 96,
    CSSPropertyBreakAfter = 97,
    CSSPropertyBreakBefore = 98,
    CSSPropertyBreakInside = 99,
    CSSPropertyBufferedRendering = 100,
    CSSPropertyCaptionSide = 101,
    CSSPropertyCaretColor = 102,
    CSSPropertyClear = 103,
    CSSPropertyClip = 104,
    CSSPropertyClipPath = 105,
    CSSPropertyClipRule = 106,
    CSSPropertyColorInterpolation = 107,
    CSSPropertyColorInterpolationFilters = 108,
    CSSPropertyColorRendering = 109,
    CSSPropertyColumnCount = 110,
    CSSPropertyColumnFill = 111,
    CSSPropertyColumnGap = 112,
    CSSPropertyColumnRule = 113,
    CSSPropertyColumnRuleColor = 114,
    CSSPropertyColumnRuleStyle = 115,
    CSSPropertyColumnRuleWidth = 116,
    CSSPropertyColumnSpan = 117,
    CSSPropertyColumnWidth = 118,
    CSSPropertyColumns = 119,
    CSSPropertyContain = 120,
    CSSPropertyContent = 121,
    CSSPropertyCounterIncrement = 122,
    CSSPropertyCounterReset = 123,
    CSSPropertyCursor = 124,
    CSSPropertyCx = 125,
    CSSPropertyCy = 126,
    CSSPropertyD = 127,
    CSSPropertyDisplay = 128,
    CSSPropertyDominantBaseline = 129,
    CSSPropertyEmptyCells = 130,
    CSSPropertyFill = 131,
    CSSPropertyFillOpacity = 132,
    CSSPropertyFillRule = 133,
    CSSPropertyFilter = 134,
    CSSPropertyFlex = 135,
    CSSPropertyFlexBasis = 136,
    CSSPropertyFlexDirection = 137,
    CSSPropertyFlexFlow = 138,
    CSSPropertyFlexGrow = 139,
    CSSPropertyFlexShrink = 140,
    CSSPropertyFlexWrap = 141,
    CSSPropertyFloat = 142,
    CSSPropertyFloodColor = 143,
    CSSPropertyFloodOpacity = 144,
    CSSPropertyFont = 145,
    CSSPropertyFontDisplay = 146,
    CSSPropertyFontVariant = 147,
    CSSPropertyGrid = 148,
    CSSPropertyGridArea = 149,
    CSSPropertyGridAutoColumns = 150,
    CSSPropertyGridAutoFlow = 151,
    CSSPropertyGridAutoRows = 152,
    CSSPropertyGridColumn = 153,
    CSSPropertyGridColumnEnd = 154,
    CSSPropertyGridColumnGap = 155,
    CSSPropertyGridColumnStart = 156,
    CSSPropertyGridGap = 157,
    CSSPropertyGridRow = 158,
    CSSPropertyGridRowEnd = 159,
    CSSPropertyGridRowGap = 160,
    CSSPropertyGridRowStart = 161,
    CSSPropertyGridTemplate = 162,
    CSSPropertyGridTemplateAreas = 163,
    CSSPropertyGridTemplateColumns = 164,
    CSSPropertyGridTemplateRows = 165,
    CSSPropertyHeight = 166,
    CSSPropertyHyphens = 167,
    CSSPropertyImageOrientation = 168,
    CSSPropertyImageRendering = 169,
    CSSPropertyInlineSize = 170,
    CSSPropertyIsolation = 171,
    CSSPropertyJustifyContent = 172,
    CSSPropertyJustifyItems = 173,
    CSSPropertyJustifySelf = 174,
    CSSPropertyLeft = 175,
    CSSPropertyLetterSpacing = 176,
    CSSPropertyLightingColor = 177,
    CSSPropertyLineBreak = 178,
    CSSPropertyLineHeight = 179,
    CSSPropertyLineHeightStep = 180,
    CSSPropertyListStyle = 181,
    CSSPropertyListStyleImage = 182,
    CSSPropertyListStylePosition = 183,
    CSSPropertyListStyleType = 184,
    CSSPropertyMargin = 185,
    CSSPropertyMarginBottom = 186,
    CSSPropertyMarginLeft = 187,
    CSSPropertyMarginRight = 188,
    CSSPropertyMarginTop = 189,
    CSSPropertyMarker = 190,
    CSSPropertyMarkerEnd = 191,
    CSSPropertyMarkerMid = 192,
    CSSPropertyMarkerStart = 193,
    CSSPropertyMask = 194,
    CSSPropertyMaskSourceType = 195,
    CSSPropertyMaskType = 196,
    CSSPropertyMaxBlockSize = 197,
    CSSPropertyMaxHeight = 198,
    CSSPropertyMaxInlineSize = 199,
    CSSPropertyMaxWidth = 200,
    CSSPropertyMaxZoom = 201,
    CSSPropertyMinBlockSize = 202,
    CSSPropertyMinHeight = 203,
    CSSPropertyMinInlineSize = 204,
    CSSPropertyMinWidth = 205,
    CSSPropertyMinZoom = 206,
    CSSPropertyMixBlendMode = 207,
    CSSPropertyObjectFit = 208,
    CSSPropertyObjectPosition = 209,
    CSSPropertyOffset = 210,
    CSSPropertyOffsetAnchor = 211,
    CSSPropertyOffsetDistance = 212,
    CSSPropertyOffsetPath = 213,
    CSSPropertyOffsetPosition = 214,
    CSSPropertyOffsetRotate = 215,
    CSSPropertyOpacity = 216,
    CSSPropertyOrder = 217,
    CSSPropertyOrientation = 218,
    CSSPropertyOrphans = 219,
    CSSPropertyOutline = 220,
    CSSPropertyOutlineColor = 221,
    CSSPropertyOutlineOffset = 222,
    CSSPropertyOutlineStyle = 223,
    CSSPropertyOutlineWidth = 224,
    CSSPropertyOverflow = 225,
    CSSPropertyOverflowAnchor = 226,
    CSSPropertyOverflowWrap = 227,
    CSSPropertyOverflowX = 228,
    CSSPropertyOverflowY = 229,
    CSSPropertyPadding = 230,
    CSSPropertyPaddingBottom = 231,
    CSSPropertyPaddingLeft = 232,
    CSSPropertyPaddingRight = 233,
    CSSPropertyPaddingTop = 234,
    CSSPropertyPage = 235,
    CSSPropertyPageBreakAfter = 236,
    CSSPropertyPageBreakBefore = 237,
    CSSPropertyPageBreakInside = 238,
    CSSPropertyPaintOrder = 239,
    CSSPropertyPerspective = 240,
    CSSPropertyPerspectiveOrigin = 241,
    CSSPropertyPlaceContent = 242,
    CSSPropertyPlaceItems = 243,
    CSSPropertyPlaceSelf = 244,
    CSSPropertyPointerEvents = 245,
    CSSPropertyPosition = 246,
    CSSPropertyQuotes = 247,
    CSSPropertyR = 248,
    CSSPropertyResize = 249,
    CSSPropertyRight = 250,
    CSSPropertyRotate = 251,
    CSSPropertyRx = 252,
    CSSPropertyRy = 253,
    CSSPropertyScale = 254,
    CSSPropertyScrollBehavior = 255,
    CSSPropertyScrollBoundaryBehavior = 256,
    CSSPropertyScrollBoundaryBehaviorX = 257,
    CSSPropertyScrollBoundaryBehaviorY = 258,
    CSSPropertyScrollPadding = 259,
    CSSPropertyScrollPaddingBlock = 260,
    CSSPropertyScrollPaddingBlockEnd = 261,
    CSSPropertyScrollPaddingBlockStart = 262,
    CSSPropertyScrollPaddingBottom = 263,
    CSSPropertyScrollPaddingInline = 264,
    CSSPropertyScrollPaddingInlineEnd = 265,
    CSSPropertyScrollPaddingInlineStart = 266,
    CSSPropertyScrollPaddingLeft = 267,
    CSSPropertyScrollPaddingRight = 268,
    CSSPropertyScrollPaddingTop = 269,
    CSSPropertyScrollSnapAlign = 270,
    CSSPropertyScrollSnapMargin = 271,
    CSSPropertyScrollSnapMarginBlock = 272,
    CSSPropertyScrollSnapMarginBlockEnd = 273,
    CSSPropertyScrollSnapMarginBlockStart = 274,
    CSSPropertyScrollSnapMarginBottom = 275,
    CSSPropertyScrollSnapMarginInline = 276,
    CSSPropertyScrollSnapMarginInlineEnd = 277,
    CSSPropertyScrollSnapMarginInlineStart = 278,
    CSSPropertyScrollSnapMarginLeft = 279,
    CSSPropertyScrollSnapMarginRight = 280,
    CSSPropertyScrollSnapMarginTop = 281,
    CSSPropertyScrollSnapStop = 282,
    CSSPropertyScrollSnapType = 283,
    CSSPropertyShapeImageThreshold = 284,
    CSSPropertyShapeMargin = 285,
    CSSPropertyShapeOutside = 286,
    CSSPropertyShapeRendering = 287,
    CSSPropertySize = 288,
    CSSPropertySpeak = 289,
    CSSPropertySrc = 290,
    CSSPropertyStopColor = 291,
    CSSPropertyStopOpacity = 292,
    CSSPropertyStroke = 293,
    CSSPropertyStrokeDasharray = 294,
    CSSPropertyStrokeDashoffset = 295,
    CSSPropertyStrokeLinecap = 296,
    CSSPropertyStrokeLinejoin = 297,
    CSSPropertyStrokeMiterlimit = 298,
    CSSPropertyStrokeOpacity = 299,
    CSSPropertyStrokeWidth = 300,
    CSSPropertyTabSize = 301,
    CSSPropertyTableLayout = 302,
    CSSPropertyTextAlign = 303,
    CSSPropertyTextAlignLast = 304,
    CSSPropertyTextAnchor = 305,
    CSSPropertyTextCombineUpright = 306,
    CSSPropertyTextDecoration = 307,
    CSSPropertyTextDecorationColor = 308,
    CSSPropertyTextDecorationLine = 309,
    CSSPropertyTextDecorationSkip = 310,
    CSSPropertyTextDecorationStyle = 311,
    CSSPropertyTextIndent = 312,
    CSSPropertyTextJustify = 313,
    CSSPropertyTextOverflow = 314,
    CSSPropertyTextShadow = 315,
    CSSPropertyTextSizeAdjust = 316,
    CSSPropertyTextTransform = 317,
    CSSPropertyTextUnderlinePosition = 318,
    CSSPropertyTop = 319,
    CSSPropertyTouchAction = 320,
    CSSPropertyTransform = 321,
    CSSPropertyTransformBox = 322,
    CSSPropertyTransformOrigin = 323,
    CSSPropertyTransformStyle = 324,
    CSSPropertyTransition = 325,
    CSSPropertyTranslate = 326,
    CSSPropertyUnicodeBidi = 327,
    CSSPropertyUnicodeRange = 328,
    CSSPropertyUserSelect = 329,
    CSSPropertyUserZoom = 330,
    CSSPropertyVectorEffect = 331,
    CSSPropertyVerticalAlign = 332,
    CSSPropertyVisibility = 333,
    CSSPropertyWebkitAppRegion = 334,
    CSSPropertyWebkitAppearance = 335,
    CSSPropertyWebkitBackgroundClip = 336,
    CSSPropertyWebkitBackgroundOrigin = 337,
    CSSPropertyWebkitBorderAfter = 338,
    CSSPropertyWebkitBorderAfterColor = 339,
    CSSPropertyWebkitBorderAfterStyle = 340,
    CSSPropertyWebkitBorderAfterWidth = 341,
    CSSPropertyWebkitBorderBefore = 342,
    CSSPropertyWebkitBorderBeforeColor = 343,
    CSSPropertyWebkitBorderBeforeStyle = 344,
    CSSPropertyWebkitBorderBeforeWidth = 345,
    CSSPropertyWebkitBorderEnd = 346,
    CSSPropertyWebkitBorderEndColor = 347,
    CSSPropertyWebkitBorderEndStyle = 348,
    CSSPropertyWebkitBorderEndWidth = 349,
    CSSPropertyWebkitBorderHorizontalSpacing = 350,
    CSSPropertyWebkitBorderImage = 351,
    CSSPropertyWebkitBorderStart = 352,
    CSSPropertyWebkitBorderStartColor = 353,
    CSSPropertyWebkitBorderStartStyle = 354,
    CSSPropertyWebkitBorderStartWidth = 355,
    CSSPropertyWebkitBorderVerticalSpacing = 356,
    CSSPropertyWebkitBoxAlign = 357,
    CSSPropertyWebkitBoxDecorationBreak = 358,
    CSSPropertyWebkitBoxDirection = 359,
    CSSPropertyWebkitBoxFlex = 360,
    CSSPropertyWebkitBoxFlexGroup = 361,
    CSSPropertyWebkitBoxLines = 362,
    CSSPropertyWebkitBoxOrdinalGroup = 363,
    CSSPropertyWebkitBoxOrient = 364,
    CSSPropertyWebkitBoxPack = 365,
    CSSPropertyWebkitBoxReflect = 366,
    CSSPropertyWebkitColumnBreakAfter = 367,
    CSSPropertyWebkitColumnBreakBefore = 368,
    CSSPropertyWebkitColumnBreakInside = 369,
    CSSPropertyWebkitFontSizeDelta = 370,
    CSSPropertyWebkitHighlight = 371,
    CSSPropertyWebkitHyphenateCharacter = 372,
    CSSPropertyWebkitLineBreak = 373,
    CSSPropertyWebkitLineClamp = 374,
    CSSPropertyWebkitLogicalHeight = 375,
    CSSPropertyWebkitLogicalWidth = 376,
    CSSPropertyWebkitMarginAfter = 377,
    CSSPropertyWebkitMarginAfterCollapse = 378,
    CSSPropertyWebkitMarginBefore = 379,
    CSSPropertyWebkitMarginBeforeCollapse = 380,
    CSSPropertyWebkitMarginBottomCollapse = 381,
    CSSPropertyWebkitMarginCollapse = 382,
    CSSPropertyWebkitMarginEnd = 383,
    CSSPropertyWebkitMarginStart = 384,
    CSSPropertyWebkitMarginTopCollapse = 385,
    CSSPropertyWebkitMask = 386,
    CSSPropertyWebkitMaskBoxImage = 387,
    CSSPropertyWebkitMaskBoxImageOutset = 388,
    CSSPropertyWebkitMaskBoxImageRepeat = 389,
    CSSPropertyWebkitMaskBoxImageSlice = 390,
    CSSPropertyWebkitMaskBoxImageSource = 391,
    CSSPropertyWebkitMaskBoxImageWidth = 392,
    CSSPropertyWebkitMaskClip = 393,
    CSSPropertyWebkitMaskComposite = 394,
    CSSPropertyWebkitMaskImage = 395,
    CSSPropertyWebkitMaskOrigin = 396,
    CSSPropertyWebkitMaskPosition = 397,
    CSSPropertyWebkitMaskPositionX = 398,
    CSSPropertyWebkitMaskPositionY = 399,
    CSSPropertyWebkitMaskRepeat = 400,
    CSSPropertyWebkitMaskRepeatX = 401,
    CSSPropertyWebkitMaskRepeatY = 402,
    CSSPropertyWebkitMaskSize = 403,
    CSSPropertyWebkitMaxLogicalHeight = 404,
    CSSPropertyWebkitMaxLogicalWidth = 405,
    CSSPropertyWebkitMinLogicalHeight = 406,
    CSSPropertyWebkitMinLogicalWidth = 407,
    CSSPropertyWebkitPaddingAfter = 408,
    CSSPropertyWebkitPaddingBefore = 409,
    CSSPropertyWebkitPaddingEnd = 410,
    CSSPropertyWebkitPaddingStart = 411,
    CSSPropertyWebkitPerspectiveOriginX = 412,
    CSSPropertyWebkitPerspectiveOriginY = 413,
    CSSPropertyWebkitPrintColorAdjust = 414,
    CSSPropertyWebkitRtlOrdering = 415,
    CSSPropertyWebkitRubyPosition = 416,
    CSSPropertyWebkitTapHighlightColor = 417,
    CSSPropertyWebkitTextCombine = 418,
    CSSPropertyWebkitTextDecorationsInEffect = 419,
    CSSPropertyWebkitTextEmphasis = 420,
    CSSPropertyWebkitTextEmphasisColor = 421,
    CSSPropertyWebkitTextEmphasisPosition = 422,
    CSSPropertyWebkitTextEmphasisStyle = 423,
    CSSPropertyWebkitTextFillColor = 424,
    CSSPropertyWebkitTextSecurity = 425,
    CSSPropertyWebkitTextStroke = 426,
    CSSPropertyWebkitTextStrokeColor = 427,
    CSSPropertyWebkitTextStrokeWidth = 428,
    CSSPropertyWebkitTransformOriginX = 429,
    CSSPropertyWebkitTransformOriginY = 430,
    CSSPropertyWebkitTransformOriginZ = 431,
    CSSPropertyWebkitUserDrag = 432,
    CSSPropertyWebkitUserModify = 433,
    CSSPropertyWhiteSpace = 434,
    CSSPropertyWidows = 435,
    CSSPropertyWidth = 436,
    CSSPropertyWillChange = 437,
    CSSPropertyWordBreak = 438,
    CSSPropertyWordSpacing = 439,
    CSSPropertyWordWrap = 440,
    CSSPropertyX = 441,
    CSSPropertyY = 442,
    CSSPropertyZIndex = 443,
    CSSPropertyAliasEpubCaptionSide = 613,
    CSSPropertyAliasEpubTextCombine = 930,
    CSSPropertyAliasEpubTextEmphasis = 932,
    CSSPropertyAliasEpubTextEmphasisColor = 933,
    CSSPropertyAliasEpubTextEmphasisStyle = 935,
    CSSPropertyAliasEpubTextOrientation = 545,
    CSSPropertyAliasEpubTextTransform = 829,
    CSSPropertyAliasEpubWordBreak = 950,
    CSSPropertyAliasEpubWritingMode = 546,
    CSSPropertyAliasWebkitAlignContent = 549,
    CSSPropertyAliasWebkitAlignItems = 550,
    CSSPropertyAliasWebkitAlignSelf = 551,
    CSSPropertyAliasWebkitAnimation = 554,
    CSSPropertyAliasWebkitAnimationDelay = 515,
    CSSPropertyAliasWebkitAnimationDirection = 516,
    CSSPropertyAliasWebkitAnimationDuration = 517,
    CSSPropertyAliasWebkitAnimationFillMode = 518,
    CSSPropertyAliasWebkitAnimationIterationCount = 519,
    CSSPropertyAliasWebkitAnimationName = 520,
    CSSPropertyAliasWebkitAnimationPlayState = 521,
    CSSPropertyAliasWebkitAnimationTimingFunction = 522,
    CSSPropertyAliasWebkitBackfaceVisibility = 556,
    CSSPropertyAliasWebkitBackgroundSize = 570,
    CSSPropertyAliasWebkitBorderBottomLeftRadius = 576,
    CSSPropertyAliasWebkitBorderBottomRightRadius = 577,
    CSSPropertyAliasWebkitBorderRadius = 592,
    CSSPropertyAliasWebkitBorderTopLeftRadius = 601,
    CSSPropertyAliasWebkitBorderTopRightRadius = 602,
    CSSPropertyAliasWebkitBoxShadow = 607,
    CSSPropertyAliasWebkitBoxSizing = 608,
    CSSPropertyAliasWebkitClipPath = 617,
    CSSPropertyAliasWebkitColumnCount = 622,
    CSSPropertyAliasWebkitColumnGap = 624,
    CSSPropertyAliasWebkitColumnRule = 625,
    CSSPropertyAliasWebkitColumnRuleColor = 626,
    CSSPropertyAliasWebkitColumnRuleStyle = 627,
    CSSPropertyAliasWebkitColumnRuleWidth = 628,
    CSSPropertyAliasWebkitColumnSpan = 629,
    CSSPropertyAliasWebkitColumnWidth = 630,
    CSSPropertyAliasWebkitColumns = 631,
    CSSPropertyAliasWebkitFilter = 646,
    CSSPropertyAliasWebkitFlex = 647,
    CSSPropertyAliasWebkitFlexBasis = 648,
    CSSPropertyAliasWebkitFlexDirection = 649,
    CSSPropertyAliasWebkitFlexFlow = 650,
    CSSPropertyAliasWebkitFlexGrow = 651,
    CSSPropertyAliasWebkitFlexShrink = 652,
    CSSPropertyAliasWebkitFlexWrap = 653,
    CSSPropertyAliasWebkitFontFeatureSettings = 530,
    CSSPropertyAliasWebkitJustifyContent = 684,
    CSSPropertyAliasWebkitOpacity = 728,
    CSSPropertyAliasWebkitOrder = 729,
    CSSPropertyAliasWebkitPerspective = 752,
    CSSPropertyAliasWebkitPerspectiveOrigin = 753,
    CSSPropertyAliasWebkitShapeImageThreshold = 796,
    CSSPropertyAliasWebkitShapeMargin = 797,
    CSSPropertyAliasWebkitShapeOutside = 798,
    CSSPropertyAliasWebkitTextSizeAdjust = 828,
    CSSPropertyAliasWebkitTransform = 833,
    CSSPropertyAliasWebkitTransformOrigin = 835,
    CSSPropertyAliasWebkitTransformStyle = 836,
    CSSPropertyAliasWebkitTransition = 837,
    CSSPropertyAliasWebkitTransitionDelay = 523,
    CSSPropertyAliasWebkitTransitionDuration = 524,
    CSSPropertyAliasWebkitTransitionProperty = 525,
    CSSPropertyAliasWebkitTransitionTimingFunction = 526,
    CSSPropertyAliasWebkitUserSelect = 841,
};

const CSSPropertyID kCSSPropertyAliasList[] = {
    static_cast<CSSPropertyID>(613), // CSSPropertyAliasEpubCaptionSide
    static_cast<CSSPropertyID>(930), // CSSPropertyAliasEpubTextCombine
    static_cast<CSSPropertyID>(932), // CSSPropertyAliasEpubTextEmphasis
    static_cast<CSSPropertyID>(933), // CSSPropertyAliasEpubTextEmphasisColor
    static_cast<CSSPropertyID>(935), // CSSPropertyAliasEpubTextEmphasisStyle
    static_cast<CSSPropertyID>(545), // CSSPropertyAliasEpubTextOrientation
    static_cast<CSSPropertyID>(829), // CSSPropertyAliasEpubTextTransform
    static_cast<CSSPropertyID>(950), // CSSPropertyAliasEpubWordBreak
    static_cast<CSSPropertyID>(546), // CSSPropertyAliasEpubWritingMode
    static_cast<CSSPropertyID>(549), // CSSPropertyAliasWebkitAlignContent
    static_cast<CSSPropertyID>(550), // CSSPropertyAliasWebkitAlignItems
    static_cast<CSSPropertyID>(551), // CSSPropertyAliasWebkitAlignSelf
    static_cast<CSSPropertyID>(554), // CSSPropertyAliasWebkitAnimation
    static_cast<CSSPropertyID>(515), // CSSPropertyAliasWebkitAnimationDelay
    static_cast<CSSPropertyID>(516), // CSSPropertyAliasWebkitAnimationDirection
    static_cast<CSSPropertyID>(517), // CSSPropertyAliasWebkitAnimationDuration
    static_cast<CSSPropertyID>(518), // CSSPropertyAliasWebkitAnimationFillMode
    static_cast<CSSPropertyID>(519), // CSSPropertyAliasWebkitAnimationIterationCount
    static_cast<CSSPropertyID>(520), // CSSPropertyAliasWebkitAnimationName
    static_cast<CSSPropertyID>(521), // CSSPropertyAliasWebkitAnimationPlayState
    static_cast<CSSPropertyID>(522), // CSSPropertyAliasWebkitAnimationTimingFunction
    static_cast<CSSPropertyID>(556), // CSSPropertyAliasWebkitBackfaceVisibility
    static_cast<CSSPropertyID>(570), // CSSPropertyAliasWebkitBackgroundSize
    static_cast<CSSPropertyID>(576), // CSSPropertyAliasWebkitBorderBottomLeftRadius
    static_cast<CSSPropertyID>(577), // CSSPropertyAliasWebkitBorderBottomRightRadius
    static_cast<CSSPropertyID>(592), // CSSPropertyAliasWebkitBorderRadius
    static_cast<CSSPropertyID>(601), // CSSPropertyAliasWebkitBorderTopLeftRadius
    static_cast<CSSPropertyID>(602), // CSSPropertyAliasWebkitBorderTopRightRadius
    static_cast<CSSPropertyID>(607), // CSSPropertyAliasWebkitBoxShadow
    static_cast<CSSPropertyID>(608), // CSSPropertyAliasWebkitBoxSizing
    static_cast<CSSPropertyID>(617), // CSSPropertyAliasWebkitClipPath
    static_cast<CSSPropertyID>(622), // CSSPropertyAliasWebkitColumnCount
    static_cast<CSSPropertyID>(624), // CSSPropertyAliasWebkitColumnGap
    static_cast<CSSPropertyID>(625), // CSSPropertyAliasWebkitColumnRule
    static_cast<CSSPropertyID>(626), // CSSPropertyAliasWebkitColumnRuleColor
    static_cast<CSSPropertyID>(627), // CSSPropertyAliasWebkitColumnRuleStyle
    static_cast<CSSPropertyID>(628), // CSSPropertyAliasWebkitColumnRuleWidth
    static_cast<CSSPropertyID>(629), // CSSPropertyAliasWebkitColumnSpan
    static_cast<CSSPropertyID>(630), // CSSPropertyAliasWebkitColumnWidth
    static_cast<CSSPropertyID>(631), // CSSPropertyAliasWebkitColumns
    static_cast<CSSPropertyID>(646), // CSSPropertyAliasWebkitFilter
    static_cast<CSSPropertyID>(647), // CSSPropertyAliasWebkitFlex
    static_cast<CSSPropertyID>(648), // CSSPropertyAliasWebkitFlexBasis
    static_cast<CSSPropertyID>(649), // CSSPropertyAliasWebkitFlexDirection
    static_cast<CSSPropertyID>(650), // CSSPropertyAliasWebkitFlexFlow
    static_cast<CSSPropertyID>(651), // CSSPropertyAliasWebkitFlexGrow
    static_cast<CSSPropertyID>(652), // CSSPropertyAliasWebkitFlexShrink
    static_cast<CSSPropertyID>(653), // CSSPropertyAliasWebkitFlexWrap
    static_cast<CSSPropertyID>(530), // CSSPropertyAliasWebkitFontFeatureSettings
    static_cast<CSSPropertyID>(684), // CSSPropertyAliasWebkitJustifyContent
    static_cast<CSSPropertyID>(728), // CSSPropertyAliasWebkitOpacity
    static_cast<CSSPropertyID>(729), // CSSPropertyAliasWebkitOrder
    static_cast<CSSPropertyID>(752), // CSSPropertyAliasWebkitPerspective
    static_cast<CSSPropertyID>(753), // CSSPropertyAliasWebkitPerspectiveOrigin
    static_cast<CSSPropertyID>(796), // CSSPropertyAliasWebkitShapeImageThreshold
    static_cast<CSSPropertyID>(797), // CSSPropertyAliasWebkitShapeMargin
    static_cast<CSSPropertyID>(798), // CSSPropertyAliasWebkitShapeOutside
    static_cast<CSSPropertyID>(828), // CSSPropertyAliasWebkitTextSizeAdjust
    static_cast<CSSPropertyID>(833), // CSSPropertyAliasWebkitTransform
    static_cast<CSSPropertyID>(835), // CSSPropertyAliasWebkitTransformOrigin
    static_cast<CSSPropertyID>(836), // CSSPropertyAliasWebkitTransformStyle
    static_cast<CSSPropertyID>(837), // CSSPropertyAliasWebkitTransition
    static_cast<CSSPropertyID>(523), // CSSPropertyAliasWebkitTransitionDelay
    static_cast<CSSPropertyID>(524), // CSSPropertyAliasWebkitTransitionDuration
    static_cast<CSSPropertyID>(525), // CSSPropertyAliasWebkitTransitionProperty
    static_cast<CSSPropertyID>(526), // CSSPropertyAliasWebkitTransitionTimingFunction
    static_cast<CSSPropertyID>(841), // CSSPropertyAliasWebkitUserSelect
};

const int firstCSSProperty = 3;
const int numCSSProperties = 441;
const int lastCSSProperty = 443;
const int lastUnresolvedCSSProperty = 950;
const int numCSSPropertyIDs = lastUnresolvedCSSProperty + 1;
const size_t maxCSSPropertyNameLength = 40;

const char* getPropertyName(CSSPropertyID);
const WTF::AtomicString& getPropertyNameAtomicString(CSSPropertyID);
WTF::String CORE_EXPORT getPropertyNameString(CSSPropertyID);
WTF::String getJSPropertyName(CSSPropertyID);

inline bool isCSSPropertyIDWithName(int id)
{
    return id >= firstCSSProperty && id <= lastUnresolvedCSSProperty;
}

inline bool isValidCSSPropertyID(CSSPropertyID id)
{
    return id != CSSPropertyInvalid;
}

inline CSSPropertyID convertToCSSPropertyID(int value)
{
    DCHECK_GE(value, CSSPropertyInvalid);
    DCHECK_LE(value, lastCSSProperty);
    return static_cast<CSSPropertyID>(value);
}

inline CSSPropertyID resolveCSSPropertyID(CSSPropertyID id)
{
    return convertToCSSPropertyID(id & ~512);
}

inline bool isPropertyAlias(CSSPropertyID id) { return id & 512; }

CSSPropertyID unresolvedCSSPropertyID(const WTF::String&);

CSSPropertyID CORE_EXPORT cssPropertyID(const WTF::String&);

} // namespace blink

#endif // CSSPropertyNames_h
